% For joint ABS and CSB optimization (eICIC)
% Only pico BSs perform CSB optimization, therefore:
% 1. If a pico BS is chosen to do the joint optimization, the CSB value of
%    the pico BS and the ABS pattern of the macro BS in the same hexagon 
%    will be optimized.
% 2. If a macro BS is chosen to do the joint optimization, only the ABS
%    pattern of the macro BS will be optimized

% The ABS patterns are pre-defined

% The value of Scheduler must be one of the followings:
% 'RR': Means round-robin
% 'PF_LTE': Means proportional fair using the one proposed in LTE reference
%           book "LTE - The UMTS Long Term Evolution: From Theory to
%           Practice", 2009, pp. 291

% Ix: Index of the station who is finding the best stragegy
% pf_window: Window size of the LTE PF scheduler
function [Station,UE] = BestResp_eICIC...
    (Station,Ix,UE,CSB_value,NumTimeSlots,NumFreqSlots,NumUE,NoiseDen,NoiseFigure,BWperRB,ABSset,NumABS,Scheduler,pf_window)

% Find the macro BS in the same hexagon, if Station(Ix) is a pico BS
if Station(Ix).Type == 1
    IxMacro = Ix;
    CSB_value = 0; % Do not perform CSB optimization if the BS is a macro BS
else
    profile = [[Station.Type]' [Station.ID]' [Station.Cluster]'];
    [nouse,IxMacro] = ismember([1 Station(Ix).MacroID 1],profile,'rows');
end

BestCSB = Station(Ix).CSB;
BestABS = Station(IxMacro).ABS;
AffectedUEix_best = cell(length(Station(Ix).AllNeighbor)+1,1);
AffectedUEix_best{1} = find([UE.attach]==Station(Ix).NameTag);
for k = 1:length(Station(Ix).AllNeighbor)
    ix = find([Station.NameTag]==Station(Ix).AllNeighbor(k));
    AffectedUEix_best{k+1} = find([UE.attach]==Station(ix).NameTag);
end
BestUtility = -Inf;
NumPRB = NumTimeSlots*NumFreqSlots;

% Check how many different user attachment patterns (w.r.t. CSB) are there
Station(Ix).CSB = 0;
[Station,UE] = update_status(Station,Ix,UE);
UE_last_CSB = Pilot_CSB_Asso(UE,NumUE,NoiseDen,NoiseFigure,BWperRB,NumTimeSlots,NumFreqSlots);
CSB_value_meaningful = zeros(1,length(CSB_value)); % Record the CSB values that lead to changes in user attachment
count = 1;
for i = 2:length(CSB_value) % Skip the first CSB value because it is zero and it has been considered
    Station(Ix).CSB = CSB_value(i);
    [Station,UE] = update_status(Station,Ix,UE);
    UE = Pilot_CSB_Asso(UE,NumUE,NoiseDen,NoiseFigure,BWperRB,NumTimeSlots,NumFreqSlots);
    if ~isequal([UE_last_CSB.attach],[UE.attach])
        count = count + 1;
        CSB_value_meaningful(count) = CSB_value(i);
        UE_last_CSB = UE;
    end  
end
if count < length(CSB_value)
    CSB_value_meaningful(count+1:end) = [];
end

% Do optimization for each of the CSB value that changes user attachment
for i = 1:length(CSB_value_meaningful)
    Station(Ix).CSB = CSB_value_meaningful(i);
    [Station,UE] = update_status(Station,Ix,UE);
    UE = Pilot_CSB_Asso(UE,NumUE,NoiseDen,NoiseFigure,BWperRB,NumTimeSlots,NumFreqSlots);
    % Reset the list of UEs that are affected by the player
    AffectedUEix = cell(length(Station(Ix).AllNeighbor)+1,1);
    AffectedUEix{1} = find([UE.attach]==Station(Ix).NameTag);
    for k = 1:length(Station(Ix).AllNeighbor)
        ix = find([Station.NameTag]==Station(Ix).AllNeighbor(k));
        AffectedUEix{k+1} = find([UE.attach]==Station(ix).NameTag);
    end
    
    % Search through all possible ABS patterns
    for j = 1:NumABS
        % Set the ABS of Station(Ix) accordingly and reschedule the users
        % who are attached to Station(Ix)
        Station(IxMacro).ABS = repmat(ABSset(j,:),[1 NumFreqSlots]);
        [Station,UE] = update_status(Station,IxMacro,UE);
        % Perform scheduling for all affected stations
        [Station,UE] = Schedule_all(Station,IxMacro,UE,NumPRB,NoiseDen,NoiseFigure,BWperRB,NumTimeSlots,NumFreqSlots,Scheduler,pf_window);
        
        % Update UE throughput
        UE(AffectedUEix{1}) = Throughput(Station(Ix).NameTag,UE(AffectedUEix{1}),NoiseDen,NoiseFigure,BWperRB,NumTimeSlots,NumFreqSlots);
        for k = 1:length(AffectedUEix)-1
            UE(AffectedUEix{k+1}) = Throughput(Station(Ix).AllNeighbor(k),UE(AffectedUEix{k+1}),NoiseDen,NoiseFigure,BWperRB,NumTimeSlots,NumFreqSlots);
        end
        
        % Find out the utility of the player, i.e., sum of log rates of users
        % who are attached to the player and the neighbours of the player
        Utility = sum(log([UE([AffectedUEix{:}]).throughput]));
        % Update the memory on setting if the current setting is better
        if Utility > BestUtility
            BestUtility = Utility;
            BestCSB = CSB_value_meaningful(i);
            BestABS = Station(IxMacro).ABS;
            AffectedUEix_best = AffectedUEix;
        end
    end  
end

% Update the settings of the current player and its mirrors
Station(Ix).CSB = BestCSB;
[Station,UE] = update_status(Station,Ix,UE);
UE = Pilot_CSB_Asso(UE,NumUE,NoiseDen,NoiseFigure,BWperRB,NumTimeSlots,NumFreqSlots);
Station(IxMacro).ABS = BestABS;
[Station,UE] = update_status(Station,IxMacro,UE);
[Station,UE] = Schedule_all(Station,IxMacro,UE,NumPRB,NoiseDen,NoiseFigure,BWperRB,NumTimeSlots,NumFreqSlots,Scheduler,pf_window);
UE(AffectedUEix_best{1}) = Throughput(Station(Ix).NameTag,UE(AffectedUEix_best{1}),NoiseDen,NoiseFigure,BWperRB,NumTimeSlots,NumFreqSlots);
for k = 1:length(AffectedUEix_best)-1
    UE(AffectedUEix_best{k+1}) = Throughput(Station(Ix).AllNeighbor(k),UE(AffectedUEix_best{k+1}),NoiseDen,NoiseFigure,BWperRB,NumTimeSlots,NumFreqSlots);
end

% to_remove = who;
% to_remove = strrep(to_remove,'Station','');
% to_remove = strrep(to_remove,'UE','');
% clear(to_remove{:})
% 
% pause(5)