% Algorithm to solve the price equilibrium problem in the following
% reference:
% J. Reijnierse and J. Potters, "On finding an envy-free pareto-optimal
% division," Mathematical Programming, vol. 83, no. 1-3, pp. 291--311,
% 1998.

% The algorithm uses the codes from SHUM Wing-Ki, Kenneth

% The algorithm is referred as cake-cutting in the reference TON paper

% Needed output: station,UE,AttachList
function [station,UE,AttachList] = PF_scheduler_Kenneth(station,UE,NumPRB,NoiseDen,NoiseFigure,BWperRB,NumTimeSlots,NumFreqSlots)

AttachList = find([UE.attach]==station.NameTag);
NumUE = length(AttachList);

if NumUE > 0 && sum(station.ABS) > 0
    [UE(AttachList)] = Throughput(station.NameTag,UE(AttachList),NoiseDen,NoiseFigure,BWperRB,NumTimeSlots,NumFreqSlots);
    throughputprofile = reshape([UE(AttachList).ThroughputPerPRB],NumPRB,NumUE)'; % Record the throughput per PRB for each user, where row i gives UE i's throughput per PRB
    % Remove PRBs that are in ABS
    [nouse,NonZeroIx] = find(throughputprofile(1,:));
    ThroughputProfile = throughputprofile(:,NonZeroIx);
    
    UE_weight = [UE(AttachList).weight] / sum([UE(AttachList).weight]); % Normalized UE weights
    
    % If the throughput of UEs on a PRB is too small, do not include that
    % PRB for cake cutting (because it could lead to error). After the cake
    % cutting, allocate those PRBs to the first user
    ToRemove_ix = find(max(ThroughputProfile)<10^-4);
    [allocation] = cake_cutting( UE_weight,ThroughputProfile(:,~ismember([1:1:length(NonZeroIx)],ToRemove_ix)) );
    for i = 1:length(ToRemove_ix)
        if ToRemove_ix(i) == 1
            allocation = [[1;zeros(NumUE-1,1)] allocation];
        elseif ToRemove_ix(i) == length(station.ABS)
            allocation = [allocation [1;zeros(NumUE-1,1)]];
        else
            allocation = [allocation(:,1:ToRemove_ix(i)-1) [1;zeros(NumUE-1,1)] allocation(:,ToRemove_ix(i):end)];
        end
    end
    
    % Do quatization to ensure every user gets at least one PRB
    % Step 1: Allocate each PRB to the UE who has the largest share
    [nouse,ix]=max(allocation,[],1);
    ix = ix + [0 cumsum(NumUE*ones(1,size(allocation,2)-1))];
    temp_Allocation = zeros(size(allocation,1),size(allocation,2));
    temp_Allocation(ix) = 1;
    copy_allocation = allocation;
    % Step 2: Make sure every user gets at least one PRB
    ZeroRowIx = find(sum(temp_Allocation,2)==0);
    FreeCol = 1:size(allocation,2);
    count = 0;
    while ~isempty(ZeroRowIx) && NumUE <= size(allocation,2)
        count = count + 1;
        % Fix this row
        RowIx = ZeroRowIx(1);
        [nouse,ColIx] = max(copy_allocation(RowIx,FreeCol));
        temp_Allocation(:,FreeCol(ColIx)) = 0;
        temp_Allocation(RowIx,FreeCol(ColIx)) = 1;
        FreeCol(ColIx) = [];
        ZeroRowIx = find(sum(temp_Allocation,2)==0);
    end
    Allocation=zeros(NumUE,NumPRB);
    Allocation(:,NonZeroIx) = temp_Allocation;
    if ~isempty(find(sum(Allocation,2)==0)) && NumUE <= size(allocation,2)
        error('New quantization method is problematic!!!')
    end
    
    % Check validity of the allocation
    if prod(Allocation(find(Allocation)) > 0) < 1 % If there are terms between 0 and 1 (not including the boundaries)
        error('Something wrong with the allocation')
    elseif ~isempty(find(Allocation(:,setdiff(1:NumPRB,NonZeroIx))))
        error('PRBs that are ABS are allocated')
    end
    % Record the allocation
    station.UEinPRB = AttachList*Allocation;
    for i = 1:NumUE
        UE(AttachList(i)).Allocation = Allocation(i,:);
    end
end